CANCELLED_APPT = "cancelled"
COMPLETE_APPT = "done"
INCOMPLETE_APPT = "incomplete"
IN_PROGRESS_APPT = "in_progress"
LAST_MONTH = "last_month"
MISSED_APPT = "missed"
NEW_APPT = "new"
NEXT_MONTH = "next_month"
ONTIME_APPT = "ontime"
SCHEDULED_APPT = "scheduled"
UNSCHEDULED_APPT = "unscheduled"
SKIPPED_APPT = "skipped"
