# Generated by Django 4.2.5 on 2023-09-25 15:43

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_appointment", "0043_alter_appointment_appt_status_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="appointment",
            name="appt_reason",
            field=models.CharField(
                choices=[
                    ("scheduled", "Scheduled visit (study)"),
                    ("unscheduled", "Routine / Unscheduled (non-study)"),
                ],
                help_text="The reason for visit from the visit report will be validated against this response. Refer to the protocol documentation for the definition of a scheduled appointment.",
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_reason",
            field=models.CharField(
                choices=[
                    ("scheduled", "Scheduled visit (study)"),
                    ("unscheduled", "Routine / Unscheduled (non-study)"),
                ],
                help_text="The reason for visit from the visit report will be validated against this response. Refer to the protocol documentation for the definition of a scheduled appointment.",
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
    ]
